//------------------------------------------------------------------------
//
// Copyright (C) 1997-2006 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module:  C++ Example
// 
// Description
//
// This is a sample C++ application which demonstrates using the Chameleon
// API to generate and send HL7 messages over an LLP connection.
//
// Author: iNTERFACEWARE
// Date:   Wed 09/09/1998 
//---------------------------------------------------------------------------
#include "CHMclientPreCompHead.h"
#pragma hdrstop

#include <cstdio>
#include <cstdlib>

#include <CHM/CHMerrorClass.h>
#include <CHM/CHMllpClient.h>
#include <iostream>
#include <string>
#include <CHM/CHMstring.h>

using namespace std;

/// This application is a simple example of using the Chameleon generated objects in a 'Client'
/// configuration.  It is intended to be a helpful starting point in understanding how to use 
/// these objects, but must be fleshed out in the following areas to provide a full solution for an
/// application HL7 interface:
/// 1) You will need to work out how your application will handle the logging of messages
///    and data, in particular how bad transactions should be logged for effective troubleshooting.
/// 2) The code will need to be modified to insert the parsed data into a database, whether
///    by using explicit database calls or by going through some type of middle tier layer that
///    your application may support.
/// 3) Some customers choose to deploy HL7 interfaces as Windows Services.  This can be done
///    easily with Chameleon.  One thing to keep in mind though is that the networking components
///    supplied with Chameleon require a Windows 'message loop'.  Often Windows Services do
///    not have a message loop - but one can be easily set up by invoking RunMessageLoop on
///    the networking component supplied with Chameleon.  See the manual for running Chameleon
///    within services.
/// 4) Definitely take the time to read the manual and go through the tutorials after looking
///    at this example and before attempting to deploy a production solution.  We do provide custom
///    on-site training and technical implementation services if you require them.
///
/// If you are looking for a much more complete 'out of the box' solution for your HL7 needs it 
/// would be a good idea to evaluate iNTERFACEWARE Iguana.  Iguana is powered by Chameleon, but
/// provides a complete run-time web-based infrastructure for your HL7 interfaces.
/// It provides advanced logging and monitoring capability which makes it a very attractive 
/// option for many customers.
/// 
/// It should be strongly emphasized that this example program is only for demonstration purposes.
/// A real HL7 interface should have some type of system of logging into log files.  Iguana, for
/// instance, has a very nice logging system that allows administrators to drill into and browse
/// processed messages and errors.  How you implement this logging is up to your requirements
/// for run-time management of your HL7 interface.

#define MENU_OPTION_DISCONNECT 0
#define MENU_OPTION_SEND_MESSAGE 1
#define MENU_OPTION_CHANGE_CONFIGURATION 2

#define NEWLINE "\r\n"
#define MAX_MESSAGE_LENGTH 200
#define GETLINE_BUFFER_SIZE 100

#define DEFAULT_HOST "localhost"
#define DEFAULT_PORT 5123

class CHMclient : public CHMllpClient
{
public:
   CHMclient(){}

   virtual ~CHMclient() {}
   virtual void OnConnected();
   virtual void OnClosed();
   virtual void OnMessageReceived(const char* Message);
   virtual void OnDataIgnored(const char* Data);
   virtual void OnErrorReceived(const char* Description, CHMint32 Code);

   PrefixTestappEngine Engine;
   CHMstring Flatwire;

private:   
   int GetInputInteger();
   void OutputLine(const char* pLine);   

   void RunMainMenu();
   void DisplayMenu();
   void DisplayConfigurationMenu();
   void DisplayMessageMenu();

};

int CHMclient::GetInputInteger()
{
   int InputChar = -1;
   cin >> InputChar;

   while (cin.fail())
   {
      cout << "Invalid input. Please enter a number." << endl;
      cin.clear();
      cin.ignore(99999999, '\n');
      cin >> InputChar;
   }

   return InputChar;
}

void CHMclient::RunMainMenu()
{
   while(true)
   {
      DisplayMenu();

      int InputChar = GetInputInteger();
      if(InputChar == MENU_OPTION_DISCONNECT)
      {
         Disconnect();
         cout << "Disconnected from server." << endl;
         break;
      }
      else if(InputChar == MENU_OPTION_SEND_MESSAGE)
      {
         try
         {
            DisplayMessageMenu();
            int MessageNumber = GetInputInteger();
  
            // Begin message generation code (comment used as an anchor for STG) 
            if((int)Engine.CountOfMessage() <= MessageNumber)
            {
               cout<<"Invalid message index. Using the default index (0)."<<endl;
               Flatwire = Engine.GenerateADT_A04();
            }
            else if(strcmp(Engine.MessageName(MessageNumber),"ADT_A04")==0)
            {
               Flatwire = Engine.GenerateADT_A04();
               cout <<"ADT_A04 message generated successfully" << endl;
            }
            // End message generation code (comment used as an anchor for STG)
                            
            cout << "Sending message: " << endl;
            cout << MakeMessagePrintable(Flatwire.c_str(), NEWLINE, MAX_MESSAGE_LENGTH)<< endl;
            TransmitMessage(Flatwire.c_str());        
            break;
         }
         catch (CHMerror& Error)
         {
            cout << endl << "An error occured while generating or sending the message:" << endl << endl;
            cout << Error.Dump().c_str() << endl;
         }
      }
      else if(InputChar == MENU_OPTION_CHANGE_CONFIGURATION)
      {
          DisplayConfigurationMenu();
          int ConfigurationNumber = GetInputInteger();
          if(ConfigurationNumber < 0 || ConfigurationNumber > (int)Engine.CountOfConfiguration())
          {
             cout << "Invalid configuration number entered. Using the default configuration (0)." << endl;
          }
          else
          {
             cout << "Current configuration set to " << ConfigurationNumber << endl;
             Engine.SetCurrentConfigurationIndex(ConfigurationNumber);
          }
      }
      else
      {
         cout << "Invalid command. Please enter a number from 0 to 2" << endl;
      }
   }
}

void CHMclient::DisplayMenu()
{
   cout << "*********************************************" << endl;
   cout << "Please select a command:" << endl;
   cout << " 0. Disconnect " << endl;
   cout << " 1. Send message" << endl;
   cout << " 2. Change configuration (currently '" << Engine.ConfigurationName(Engine.CurrentConfigurationIndex()) << "')" << endl;
}

void CHMclient::DisplayConfigurationMenu()
{
   cout << "*********************************************"<< endl;
   cout << "Please select a configuration: " << endl;

   for (int ConfigIndex = 0; ConfigIndex < (int)Engine.CountOfConfiguration(); ConfigIndex++) 
   {
      cout << " " << ConfigIndex << ". " << Engine.ConfigurationName(ConfigIndex) << endl;
   }
}

void CHMclient::DisplayMessageMenu()
{
   cout << "*********************************************" << endl;
   cout << "Please select a message to send:" << endl;

   for(int CountOfMessage = 0; CountOfMessage < (int)Engine.CountOfMessage(); CountOfMessage++)
   {
      cout << " " << CountOfMessage  << ". " << Engine.MessageName(CountOfMessage) << endl;
   }
  
   cout << "*********************************************" << endl; 
}

// This event fires when the Client connects to the Server.
// The connected event is very important.  One common mistake
// that developers often make when new to socket programming
// is to try and invoke the Connect method and then immediately
// send data via the socket.  This is not possible to do until the
// socket is actually connected.  Instead,  you should invoke the
// Connect method and then wait for this event before sending
// data.
void CHMclient::OnConnected()
{
   cout << "Connection to server established." << endl;
   cout << endl;
   
   RunMainMenu();
}

// This event fires when the Client is disconnected from the Server.
void CHMclient::OnClosed()
{
   cout <<  "Connection to server closed." << endl;
   //need to stop the message loop when client's disconnected
   CHMllpClient::StopMessageLoop();
}

// This event fires when a message is received.  Usually the message received 
// is an ACK message.  If, however, you are implementing a lab query interface 
// it might be a lab result message.  For this example we are assuming it is an ACK message and
// just printing it.
void CHMclient::OnMessageReceived
(
   const char* Message 
)
{
   cout << "Received an ACK:" << endl;
   CHMstring MessageString = MakeMessagePrintable(Message, NEWLINE, MAX_MESSAGE_LENGTH);
   cout <<  MessageString.c_str() << endl;

   RunMainMenu();
}

// This event fires when we receive a TCP/IP error during connection.
// This will always terminate the connection, so your real application will 
// need logic to determine if and when to re-connect to the other host.
void CHMclient::OnErrorReceived
(
   const char* Description,
   CHMint32 Code
)
{
   cout << Description << endl;
   // We do not need to disconnect since the connection will already have
   // been closed.
 
   // In this example we stop the message loop when an error is received 
   // to break out of the message loop and end this program.  
   // A real outgoing HL7 client program would probably not do this.
   CHMllpClient::StopMessageLoop();
}

// This event fires when we receive data that is not using the Lower Layer Protocol (LLP).
// LLP is the standard protocol for sending HL7 messages, and is used by Chameleon.  See the manual
// for more details.
void CHMclient::OnDataIgnored
(
   const char* Data
)
{
   CHMstring Output = "Received ignored data:";
   Output += Data;
   cout << Output.c_str() << endl;

}

int main()
{  
   try
   {
      CHMclient Client;

      cout << "Please verify that the listener object is started when connecting the client." << endl; 
      cout << "Starting up client..." << endl;

      // In order to make the engine usable we must load the vmd file.  This file
      // contains the information about the structure of the messages, tables, and data
      // mappings that were defined in the Chameleon GUI.  Client.Engine is a generated 
      // Engine which has an interface based on the messages and tables defined in the vmd file.
      // 
      // Also keep in mind that this example is using the static stubcode 
      // interface of Chameleon. For advanced applications it is possible to 
      // use a dynamic interface to populate messages.  See the manual for 
      // more details.
      Client.Engine.Load("test_data.vmd");
      Client.Engine.SetLog(new CoutOutputLog());

      cout << "Loaded message definition file 'test_data.vmd'." << endl;

      cout << "Enter the host to connect to (default is '" << DEFAULT_HOST << "'): " << endl;
      char Input[GETLINE_BUFFER_SIZE];
      CHMstring Host = DEFAULT_HOST;
      cin.getline(Input, GETLINE_BUFFER_SIZE);      
      if(Input[0] != 0)
      {
         Host = Input;
      }

      cout << "Enter the port number to connect to (default is '" << DEFAULT_PORT << "'): " << endl;
      cin.getline(Input, GETLINE_BUFFER_SIZE);

      int Port = DEFAULT_PORT;
      if(Input[0] != 0)
      {
         Port = atoi(Input);
      }

      if(!Client.IsConnected())
      {
         // Connect to the listener
         cout << "Connecting to port " << Port << " on " << Host.c_str() << "..." << endl;
         try
         {
            Client.Connect(Host.c_str(), Port);
         }
         catch(CHMerror& Error)
         {
            cout << "Error when connecting to server " << Error.Description() << endl;
            return EXIT_FAILURE;
         }
      
      }

      // Need to call runMessageLoop() to start up the client. Note that this is a blocking
      // call, so the user input code is put into the event itself.
      CHMllpClient::RunMessageLoop();

      cout << endl << "*********************************************" << endl << endl;
   }
   catch (CHMerror& Error)
   {
      cout << Error.Dump().c_str() << endl;
      cout << endl << "*********************************************" << endl << endl;
   }

   return EXIT_SUCCESS;
}
